﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace Paketbote
{
    public partial class frm_Paketbote : Form
    {

        Gameboard gb;
        Gameboard gb2;
        AboutBox frm_About;
        public static Graphics form_Graphics;
        public static Brush brush_white;
        public static Brush brush_green;
        public static TextureBrush brush_pkg;
        private bool mp = false;
        private int cnt_safety_1;
        private int cnt_safety_2;
        private Random rnd;
        private Queue<spawn_pkg> q_spawn_1;
        private Queue<spawn_pkg> q_spawn_2;
        private bool quit;
        private bool initialized = false;

        //structure for remembering the spawned packages in multiplayer
        public struct spawn_pkg
        {
            public enum_type type;
            public int col;
        }

        /// <summary>
        /// initializes the game-form
        /// </summary>
        /// <param name="mp">true for 2 players</param>
        /// <param name="rows">number of rows</param>
        /// <param name="cols">number of columns</param>
        /// <param name="speed">initial speed value</param>
        public frm_Paketbote(bool mp, int rows, int cols,int speed)
        {

            //Initializing necessary components, resizing window and more

            InitializeComponent();
            rnd = new Random();
            q_spawn_1 = new Queue<spawn_pkg>();
            q_spawn_2 = new Queue<spawn_pkg>();
            quit = true;
            this.mp = mp;           
            brush_white = new SolidBrush(Color.White);
            brush_green = new SolidBrush(Color.Green);
            brush_pkg = new TextureBrush(Properties.Resources.icon_pkg);
            timer_1.Interval = speed;
            timer_2.Interval = speed;
            gb = new Gameboard(rows,cols, new Point(10, 60));
            lbl_score_1.Location = new Point(20,lbl_score_1.Location.Y);
            lbl_score_val_1.Location = new Point(lbl_score_1.Location.X + lbl_score_1.Width + 10, lbl_score_1.Location.Y);
            lbl_score_2.Location = new Point(70+cols*40, lbl_score_1.Location.Y);
            lbl_score_val_2.Location = new Point(lbl_score_2.Location.X + lbl_score_2.Width + 10, lbl_score_1.Location.Y);
            this.Size = new Size(35 + cols *40,110 + rows*40);
            if (mp)
            {      
                gb2 = new Gameboard(rows, cols, new Point(10+cols*40+50, 60));
                this.Size = new Size(85 + 2 * cols*40, 110 + rows*40);
            }
            form_Graphics = this.CreateGraphics();
            form_Graphics.InterpolationMode = System.Drawing.Drawing2D.InterpolationMode.High;
        }

        /// <summary>
        /// initializes/resets gameboard and starts the game
        /// </summary>
        private void start_Game()
        {
            form_Graphics.Clear(Color.White);

            gb.clear_Gameboard();
            gb.draw_full_Gameboard();
            gb.draw_Border();
            gb.stop_Key_Input = false;
            timer_1.Start();
            
            if (mp)
            {
                gb2.clear_Gameboard();
                gb2.draw_full_Gameboard();
                gb2.draw_Border();
                gb2.stop_Key_Input = false;
                timer_2.Start();                             
            }         
        }

        /// <summary>
        /// stops the game and prints message
        /// </summary>
        public void GAME_OVER(int gb_sender)
        {
            initialized = false;
            timer_1.Stop();
            gb.stop_Key_Input = true;
            if (mp)
            {
                timer_2.Stop();
                gb2.stop_Key_Input = true;

                if (gb_sender==1)
                {
                    MessageBox.Show("Spieler 2 hat gewonnen!", "GAME OVER", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }
                else
                {
                    MessageBox.Show("Spieler 1 hat gewonnen!", "GAME OVER", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
                }
            }
            else
            {
                MessageBox.Show("Das Spiel ist vorbei!", "GAME OVER", MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
            }
            
        }

        /// <summary>
        /// generates a random spawn-package-structure
        /// </summary>
        /// <returns>a random spawn-package-structure</returns>
        public spawn_pkg generate_spawn_pkg()
        {
            spawn_pkg rnd_pkg = new spawn_pkg();
            int col = rnd.Next(2, gb.cols - 2);
            int i_type = rnd.Next(2, 9);
            enum_type e_type = enum_type.T;
            e_type = (enum_type)i_type;
            rnd_pkg.col = col;
            rnd_pkg.type = e_type;
            return rnd_pkg;
        }

        /// <summary>
        /// spawns the same packages on both boards
        /// </summary>
        /// <param name="gb_sel">gameboard id on which package should be spawned; 0 for both</param>
        public void spawn_Package_RND(int gb_sel = 0)
        {
            spawn_pkg curr_pkg;

            if (gb_sel == 0)
            {
                curr_pkg = generate_spawn_pkg();
                gb.spawn_Package(curr_pkg.type, curr_pkg.col);
                if (mp)
                {
                    gb2.spawn_Package(curr_pkg.type, curr_pkg.col);
                }
            }
            else if (gb_sel == 1)
            {
                if (q_spawn_1.Count!=0)
                {
                    curr_pkg = q_spawn_1.Dequeue();
                    gb.spawn_Package(curr_pkg.type, curr_pkg.col);
                }
                else
                {
                    curr_pkg = generate_spawn_pkg();
                    q_spawn_2.Enqueue(curr_pkg);
                    gb.spawn_Package(curr_pkg.type, curr_pkg.col);
                }             
            }
            else if (gb_sel == 2)
            {
                if (q_spawn_2.Count != 0)
                {
                    curr_pkg = q_spawn_2.Dequeue();
                    gb2.spawn_Package(curr_pkg.type, curr_pkg.col);
                }
                else
                {
                    curr_pkg = generate_spawn_pkg();
                    q_spawn_1.Enqueue(curr_pkg);
                    gb2.spawn_Package(curr_pkg.type, curr_pkg.col);
                }
            }
            else{ }
        }

        /// <summary>
        /// timer which moves the packages down and calculates the new score
        /// </summary>
        private void timer_1_Tick(object sender, EventArgs e)
        {
            int return_val;
            //stop key-input while moving blocks down to prevent displacement
            gb.stop_Key_Input = true;
            return_val = gb.calculate_next_Gamestate("down");
            if (return_val == 1)
            {
                //disable key-input once package has reached bottom until fully placed
                if (gb.update_next_Gamestate("down"))
                {
                    gb.stop_Key_Input = true;
                }
                else
                {
                    gb.stop_Key_Input = false;
                }
            }
            else if (return_val == 2)
            {
                GAME_OVER(1);
                return;
            }
            else
            {
                gb.score += gb.stop_moving_Blocks().points;
                gb.draw_full_Gameboard();
                spawn_Package_RND(1);
                lbl_score_val_1.Text = Convert.ToString(gb.score);
            }
            gb.draw_partial_Gameboard();
        }

        /// <summary>
        /// timer (MP) which moves the packages down and calculates the new score
        /// </summary>
        private void timer_2_Tick(object sender, EventArgs e)
        {
            int return_val;
            //stop key-input while moving blocks down to prevent displacement
            gb2.stop_Key_Input = true;
            return_val = gb2.calculate_next_Gamestate("down");
            if (return_val == 1)
            {
                //disable key-input once package has reached bottom until fully placed
                if (gb2.update_next_Gamestate("down"))
                {
                    gb2.stop_Key_Input = true;
                }
                else
                {
                    gb2.stop_Key_Input = false;
                }
            }
            else if (return_val == 2)
            {
                GAME_OVER(2);
                return;
            }
            else
            {
                gb2.score += gb2.stop_moving_Blocks().points;
                gb2.draw_full_Gameboard();
                spawn_Package_RND(2);
                lbl_score_val_2.Text = Convert.ToString(gb2.score);
            }
            gb2.draw_partial_Gameboard();
        }

        /// <summary>
        /// starts a new game
        /// </summary>
        private void tsm_new_Click(object sender, EventArgs e)
        {
            gb.score = 0;
            lbl_score_val_1.Text = "0";
            if (mp)
            {
                gb2.score = 0;
                lbl_score_val_2.Text = "0";
            }    
            if (timer_1.Enabled || timer_2.Enabled)
            {
                timer_1.Stop();
                timer_2.Stop();
            }
            q_spawn_1.Clear();
            q_spawn_2.Clear();
            start_Game();
            spawn_Package_RND();
            initialized = true;
        }
        /// <summary>
        /// pauses or unpauses game
        /// </summary>
        private void tsm_pause_Click(object sender, EventArgs e)
        {
            if (initialized)
            {
                if (timer_1.Enabled || timer_2.Enabled)
                {
                    gb.stop_Key_Input = true;
                    timer_1.Stop();
                    if (mp)
                    {
                        gb2.stop_Key_Input = true;
                        timer_2.Stop();
                    }

                    tsm_pause.Text = "Fortsetzen";
                    tsm_pause.Image = Properties.Resources.icon_play;
                }
                else
                {
                    timer_1.Start();
                    gb.stop_Key_Input = false;

                    if (mp)
                    {
                        timer_2.Start();
                        gb2.stop_Key_Input = false;
                    }
                    tsm_pause.Text = "Pausieren";
                    tsm_pause.Image = Properties.Resources.icon_pause;
                }
            }
        }

        /// <summary>
        /// quits the game
        /// </summary>
        private void tsm_quit_Click(object sender, EventArgs e)
        {
            if (timer_1.Enabled || timer_2.Enabled)
            {
                timer_1.Stop();
                timer_2.Stop();
            }
            form_Graphics.Dispose();
            Application.Exit();
        }

        /// <summary>
        /// waits for key-input and checks if move in direction is possible and executes action accordingly
        /// </summary>
        protected override bool ProcessCmdKey(ref Message msg, Keys keyData)
        {
            if (keyData == Keys.P)
            {
                tsm_pause.PerformClick();
            }

            if (!gb.stop_Key_Input)
            {
                //move package left if possible
                if (keyData == Properties.Settings.Default.key_sp_left)
                {
                    if (gb.calculate_next_Gamestate("left") == 1)
                    {
                        gb.update_next_Gamestate("left");
                    }
                    gb.draw_partial_Gameboard();
                }
                //move package right if possible
                else if (keyData == Properties.Settings.Default.key_sp_right)
                {
                    if (gb.calculate_next_Gamestate("right") == 1)
                    {
                        gb.update_next_Gamestate("right");
                    }
                    gb.draw_partial_Gameboard();
                }
                //drop the package with disabling key-input and moving down until not possible anymore
                else if (keyData == Properties.Settings.Default.key_sp_down && Properties.Settings.Default.mode_classic_sp)
                {
                    cnt_safety_1 = 0;
                    gb.stop_Key_Input = true;
                    timer_1.Stop();
                    while (gb.calculate_next_Gamestate("down") == 1 && cnt_safety_1 < gb.rows)
                    {
                        gb.update_next_Gamestate("down");
                        cnt_safety_1++;
                    }
                    gb.score += gb.stop_moving_Blocks().points;
                    lbl_score_val_1.Text = Convert.ToString(gb.score);
                    gb.draw_full_Gameboard();
                    spawn_Package_RND(1);
                    timer_1.Start();
                    gb.stop_Key_Input = true;
                }
                //rotate package left if possible
                else if (keyData == Properties.Settings.Default.key_sp_down && !Properties.Settings.Default.mode_classic_sp)
                {
                    if (gb.calculate_next_Gamestate("rotate_left") == 1)
                    {
                        gb.update_next_Gamestate("rotate_left");
                    }
                    gb.draw_partial_Gameboard();
                }
                //roate package right if possible
                else if (keyData == Properties.Settings.Default.key_sp_up)
                {
                    if (gb.calculate_next_Gamestate("rotate_right") == 1)
                    {
                        gb.update_next_Gamestate("rotate_right");
                    }
                    gb.draw_partial_Gameboard();
                }
                else { }
            }
            if (mp)
            {
                if (!gb2.stop_Key_Input)
                {
                    //move package left if possible
                    if (keyData == Properties.Settings.Default.key_mp_left)
                    {
                        if (gb2.calculate_next_Gamestate("left") == 1)
                        {
                            gb2.update_next_Gamestate("left");
                        }
                        gb2.draw_partial_Gameboard();
                    }
                    //move package right if possible
                    else if (keyData == Properties.Settings.Default.key_mp_right)
                    {
                        if (gb2.calculate_next_Gamestate("right") == 1)
                        {
                            gb2.update_next_Gamestate("right");
                        }
                        gb2.draw_partial_Gameboard();
                    }
                    //drop the package with disabling key-input and moving down until not possible anymore
                    else if (keyData == Properties.Settings.Default.key_mp_down && Properties.Settings.Default.mode_classic_mp)
                    {
                        cnt_safety_2 = 0;
                        gb2.stop_Key_Input = true;
                        timer_2.Stop();
                        while (gb2.calculate_next_Gamestate("down") == 1 && cnt_safety_2 < gb2.rows)
                        {
                            gb2.update_next_Gamestate("down");
                            cnt_safety_2++;
                        }
                        gb2.score += gb2.stop_moving_Blocks().points;
                        lbl_score_val_2.Text = Convert.ToString(gb2.score);
                        gb2.draw_full_Gameboard();
                        spawn_Package_RND(2);
                        timer_2.Start();
                        gb2.stop_Key_Input = true;
                    }
                    //rotate package left if possible
                    else if (keyData == Properties.Settings.Default.key_mp_down && !Properties.Settings.Default.mode_classic_mp)
                    {
                        if (gb2.calculate_next_Gamestate("rotate_left") == 1)
                        {
                            gb2.update_next_Gamestate("rotate_left");
                        }
                        gb2.draw_partial_Gameboard();
                    }
                    //rotate package right if possible
                    else if (keyData == Properties.Settings.Default.key_mp_up)
                    {
                        if (gb2.calculate_next_Gamestate("rotate_right") == 1)
                        {
                            gb2.update_next_Gamestate("rotate_right");
                        }
                        gb2.draw_partial_Gameboard();
                    }
                    else { }
                }
            }
            return base.ProcessCmdKey(ref msg, keyData);
        }

        //closes game-window and reopens main menu
        private void tsm_menu_Click(object sender, EventArgs e)
        {
            quit = false;
            this.Close();
            Program.frm_load.Show();
        }

        //disable timers and close whole application
        private void frm_Paketbote_FormClosed(object sender, FormClosedEventArgs e)
        {
            if (timer_1.Enabled || timer_2.Enabled)
            {
                timer_1.Stop();
                timer_2.Stop();
            }
            form_Graphics.Dispose();
            if (quit)
            {
                Application.Exit();
            }
        }

        //open about-box
        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            frm_About = new AboutBox();
            frm_About.Location = new Point(this.Location.X, this.Location.Y);
            frm_About.Show();
        }
    }
}
